@@ -80,7 +80,7 @@ create_tmp_file (const char *template,
 
 # ifdef USE_ONLY_8DOT3
   /* Here is another Windoze bug?:
-   * you cant rename("pubring.kbx.tmp", "pubring.kbx");
+   * you can't rename("pubring.kbx.tmp", "pubring.kbx");
    * but	rename("pubring.kbx.tmp", "pubring.aaa");
    * works.  So we replace ".kbx" by ".kb_" or ".k__".  Note that we
    * can't use ".bak" and ".tmp", because these suffixes are used by
@@ -603,7 +603,7 @@ keybox_set_flags (KEYBOX_HANDLE hd, int what, int idx, unsigned int value)
 
   ec = 0;
   if (fseeko (fp, off, SEEK_SET))
-    ec = gpg_error_from_syserror ();
+    ec = gpg_err_code_from_syserror ();
   else
     {
       unsigned char tmp[4];
